#ifndef MSP_GL_SL_SYNTAX_H_
#define MSP_GL_SL_SYNTAX_H_

#include <list>
#include <map>
#include <set>
#include <string>
#include <vector>
#include <msp/core/inttypes.h>
#include <msp/core/refptr.h>
#include <msp/core/variant.h>
#include "features.h"
#include "glsl_error.h"
#include "sourcemap.h"

#pragma push_macro("interface")
#undef interface

namespace Msp {
namespace GL {
namespace SL {

struct Operator
{
	enum Type
	{
		NO_OPERATOR,
		BINARY,
		PREFIX,
		POSTFIX,
		TERNARY
	};

	enum Associativity
	{
		LEFT_TO_RIGHT,
		RIGHT_TO_LEFT,
		ASSOCIATIVE
	};

	char token[4];
	char token2[2];
	UInt8 precedence;
	Type type;
	Associativity assoc;

	static const Operator operators[];

	static const Operator &get_operator(const std::string &, Type);
};

enum
{
	INTERNAL_SOURCE = -2,
	BUILTIN_SOURCE = -1,
	GENERATED_SOURCE = 0
};

struct NodeVisitor;

struct Node
{
	int source;
	unsigned line;

	Node(): source(GENERATED_SOURCE), line(1) { }
	Node(const Node &n): source(n.source), line(n.line) { }
private:
	Node &operator=(const Node &);
public:
	virtual ~Node() { }

	virtual Node *clone() const = 0;
	virtual void visit(NodeVisitor &) = 0;
};

template<typename T>
class NodePtr: public RefPtr<T>
{
public:
	NodePtr() { }
	NodePtr(T *p): RefPtr<T>(p) { }
	NodePtr(const NodePtr &p): RefPtr<T>(p ? p->clone() : 0) { }
	NodePtr &operator=(const NodePtr &p) { RefPtr<T>::operator=(p); return *this; }

	template<typename U>
	NodePtr(const RefPtr<U> &p): RefPtr<T>(p) { }

	template<typename U>
	NodePtr(const NodePtr<U> &p): RefPtr<T>(p ? p->clone() : 0) { }
};

template<typename C>
class NodeContainer: public C
{
public:
	NodeContainer() { }
	NodeContainer(const NodeContainer &);

	void push_back_nocopy(const typename C::value_type &v)
	{ C::push_back(0); C::back() = v; }
};

template<typename T>
class NodeList: public NodeContainer<std::list<RefPtr<T> > >
{ };

template<typename T>
class NodeArray: public NodeContainer<std::vector<RefPtr<T> > >
{ };

struct TypeDeclaration;
struct VariableDeclaration;
struct InterfaceBlock;
struct FunctionDeclaration;

struct Statement: Node
{
	virtual Statement *clone() const = 0;
};

struct Block: Node
{
	NodeList<Statement> body;
	bool use_braces;

	std::map<std::string, VariableDeclaration *> variables;
	Block *parent;

	Block();
	Block(const Block &);

	virtual Block *clone() const { return new Block(*this); }
	virtual void visit(NodeVisitor &);
};

struct Expression: Node
{
	const Operator *oper;

	TypeDeclaration *type;
	bool lvalue;

	Expression();

	virtual Expression *clone() const = 0;
};

struct Literal: Expression
{
	std::string token;
	Variant value;

	virtual Literal *clone() const { return new Literal(*this); }
	virtual void visit(NodeVisitor &);
};

struct VariableReference: Expression
{
	std::string name;

	VariableDeclaration *declaration;

	VariableReference();
	VariableReference(const VariableReference &);

	virtual VariableReference *clone() const { return new VariableReference(*this); }
	virtual void visit(NodeVisitor &);
};

struct InterfaceBlockReference: Expression
{
	std::string name;

	InterfaceBlock *declaration;

	InterfaceBlockReference();
	InterfaceBlockReference(const InterfaceBlockReference &);

	virtual InterfaceBlockReference *clone() const { return new InterfaceBlockReference(*this); }
	virtual void visit(NodeVisitor &);
};

struct MemberAccess: Expression
{
	NodePtr<Expression> left;
	std::string member;

	VariableDeclaration *declaration;
	int index;

	MemberAccess();
	MemberAccess(const MemberAccess &);

	virtual MemberAccess *clone() const { return new MemberAccess(*this); }
	virtual void visit(NodeVisitor &);
};

struct Swizzle: Expression
{
	NodePtr<Expression> left;
	std::string component_group;
	unsigned count;
	UInt8 components[4];

	Swizzle();

	virtual Swizzle *clone() const { return new Swizzle(*this); }
	virtual void visit(NodeVisitor &);
};

struct UnaryExpression: Expression
{
	NodePtr<Expression> expression;

	virtual UnaryExpression *clone() const { return new UnaryExpression(*this); }
	virtual void visit(NodeVisitor &);
};

struct BinaryExpression: Expression
{
	NodePtr<Expression> left;
	NodePtr<Expression> right;

	virtual BinaryExpression *clone() const { return new BinaryExpression(*this); }
	virtual void visit(NodeVisitor &);
};

struct Assignment: BinaryExpression
{
	struct Target
	{
		enum ChainType
		{
			MEMBER = 0x40,
			SWIZZLE = 0x80,
			ARRAY = 0xC0
		};

		Statement *declaration;
		Msp::UInt8 chain_len;
		Msp::UInt8 chain[7];

		Target(Statement * = 0);

		bool operator<(const Target &) const;
	};

	bool self_referencing;

	Target target;

	Assignment();
	Assignment(const Assignment &);

	virtual Assignment *clone() const { return new Assignment(*this); }
	virtual void visit(NodeVisitor &);
};

struct TernaryExpression: Expression
{
	NodePtr<Expression> condition;
	NodePtr<Expression> true_expr;
	NodePtr<Expression> false_expr;

	virtual TernaryExpression *clone() const { return new TernaryExpression(*this); }
	virtual void visit(NodeVisitor &);
};

struct FunctionCall: Expression
{
	std::string name;
	bool constructor;
	NodeArray<Expression> arguments;

	FunctionDeclaration *declaration;

	FunctionCall();
	FunctionCall(const FunctionCall &);

	virtual FunctionCall *clone() const { return new FunctionCall(*this); }
	virtual void visit(NodeVisitor &);
};

struct ExpressionStatement: Statement
{
	NodePtr<Expression> expression;

	virtual ExpressionStatement *clone() const { return new ExpressionStatement(*this); }
	virtual void visit(NodeVisitor &);
};

struct Import: Statement
{
	std::string module;

	virtual Import *clone() const { return new Import(*this); }
	virtual void visit(NodeVisitor &);
};

struct Precision: Statement
{
	std::string precision;
	std::string type;

	virtual Precision *clone() const { return new Precision(*this); }
	virtual void visit(NodeVisitor &);
};

struct Layout: Node
{
	struct Qualifier
	{
		std::string name;
		bool has_value;
		int value;

		Qualifier(const std::string &n = std::string()): name(n), has_value(false), value(0) { }
		Qualifier(const std::string &n, int v): name(n), has_value(true), value(v) { }
	};

	std::vector<Qualifier> qualifiers;

	virtual Layout *clone() const { return new Layout(*this); }
	virtual void visit(NodeVisitor &);
};

struct InterfaceLayout: Statement
{
	std::string interface;
	Layout layout;

	virtual InterfaceLayout *clone() const { return new InterfaceLayout(*this); }
	virtual void visit(NodeVisitor &);
};

struct TypeDeclaration: Statement
{
	std::string name;

	virtual TypeDeclaration *clone() const = 0;
};

struct BasicTypeDeclaration: TypeDeclaration
{
	enum Kind
	{
		ALIAS,
		VOID,
		BOOL,
		INT,
		FLOAT,
		VECTOR,
		MATRIX,
		ARRAY
	};

	Kind kind;
	unsigned size;
	std::string base;

	TypeDeclaration *base_type;

	BasicTypeDeclaration();
	BasicTypeDeclaration(const BasicTypeDeclaration &);

	virtual BasicTypeDeclaration *clone() const { return new BasicTypeDeclaration(*this); }
	virtual void visit(NodeVisitor &);
};

struct ImageTypeDeclaration: TypeDeclaration
{
	enum Dimensions
	{
		ONE = 1,
		TWO,
		THREE,
		CUBE
	};

	Dimensions dimensions;
	bool array;
	bool sampled;
	bool shadow;
	std::string base;

	TypeDeclaration *base_type;

	ImageTypeDeclaration();

	virtual ImageTypeDeclaration *clone() const { return new ImageTypeDeclaration(*this); }
	virtual void visit(NodeVisitor &);
};

struct StructDeclaration: TypeDeclaration
{
	Block members;

	InterfaceBlock *interface_block;

	StructDeclaration();
	StructDeclaration(const StructDeclaration &);
	~StructDeclaration();

	virtual StructDeclaration *clone() const { return new StructDeclaration(*this); }
	virtual void visit(NodeVisitor &);
};

struct VariableDeclaration: Statement
{
	NodePtr<Layout> layout;
	bool constant;
	std::string sampling;
	std::string interpolation;
	std::string interface;
	std::string precision;
	std::string type;
	std::string name;
	bool array;
	NodePtr<Expression> array_size;
	NodePtr<Expression> init_expression;

	TypeDeclaration *type_declaration;
	VariableDeclaration *linked_declaration;

	VariableDeclaration();
	VariableDeclaration(const VariableDeclaration &);
	~VariableDeclaration();

	virtual VariableDeclaration *clone() const { return new VariableDeclaration(*this); }
	virtual void visit(NodeVisitor &);
};

struct InterfaceBlock: Statement
{
	NodePtr<Layout> layout;
	std::string interface;
	std::string block_name;
	NodePtr<Block> members;
	std::string instance_name;
	bool array;

	/* An interface block's ultimate base type is always a struct.  The
	immediate type may be either that same struct or an array of it. */
	TypeDeclaration *type_declaration;
	StructDeclaration *struct_declaration;
	InterfaceBlock *linked_block;

	InterfaceBlock();
	InterfaceBlock(const InterfaceBlock &);
	~InterfaceBlock();

	virtual InterfaceBlock *clone() const { return new InterfaceBlock(*this); }
	virtual void visit(NodeVisitor &);
};

struct FunctionDeclaration: Statement
{
	std::string return_type;
	std::string name;
	NodeArray<VariableDeclaration> parameters;
	bool virtua;
	bool overrd;
	Block body;

	std::string signature;
	FunctionDeclaration *definition;
	TypeDeclaration *return_type_declaration;

	FunctionDeclaration();
	FunctionDeclaration(const FunctionDeclaration &);

	virtual FunctionDeclaration *clone() const { return new FunctionDeclaration(*this); }
	virtual void visit(NodeVisitor &);
};

struct Conditional: Statement
{
	NodePtr<Expression> condition;
	Block body;
	Block else_body;

	virtual Conditional *clone() const { return new Conditional(*this); }
	virtual void visit(NodeVisitor &);
};

struct Iteration: Statement
{
	NodePtr<Statement> init_statement;
	NodePtr<Expression> condition;
	NodePtr<Expression> loop_expression;
	Block body;

	virtual Iteration *clone() const { return new Iteration(*this); }
	virtual void visit(NodeVisitor &);
};

struct Passthrough: Statement
{
	NodePtr<Expression> subscript;

	virtual Passthrough *clone() const { return new Passthrough(*this); }
	virtual void visit(NodeVisitor &);
};

struct Return: Statement
{
	NodePtr<Expression> expression;

	virtual Return *clone() const { return new Return(*this); }
	virtual void visit(NodeVisitor &);
};

struct Jump: Statement
{
	std::string keyword;

	virtual Jump *clone() const { return new Jump(*this); }
	virtual void visit(NodeVisitor &);
};

struct Stage
{
	enum Type
	{
		SHARED,
		VERTEX,
		GEOMETRY,
		FRAGMENT
	};

	Type type;
	Stage *previous;
	Block content;
	std::map<std::string, TypeDeclaration *> types;
	std::map<std::string, InterfaceBlock *> interface_blocks;
	std::map<std::string, FunctionDeclaration *> functions;
	std::map<std::string, unsigned> locations;
	std::map<std::string, unsigned> texture_bindings;
	std::map<std::string, unsigned> uniform_block_bindings;
	Features required_features;
	std::vector<Diagnostic> diagnostics;

	Stage(Type);

	static const char *get_stage_name(Type);
};

struct Module
{
	SourceMap source_map;
	Stage shared;
	std::list<Stage> stages;

	Module();
};

std::string get_unused_variable_name(const Block &, const std::string &);

int get_layout_value(const Layout &, const std::string &, int = -1);
void add_to_chain(Assignment::Target &, Assignment::Target::ChainType, unsigned);

} // namespace SL
} // namespace GL
} // namespace Msp

#pragma pop_macro("interface")

#endif
