#include "basicmaterial.h"

using namespace std;

namespace Msp {
namespace GL {

const Tag BasicMaterial::texture_tags[] =
{
	Tag("diffuse_map"),
	Tag("specular_map"),
	Tag("shininess_map"),
	Tag("normal_map"),
	Tag("emission_map"),
	Tag("reflectivity_map"),
	Tag()
};

BasicMaterial::BasicMaterial():
	receive_shadows(false)
{
	set_diffuse(Color(1.0f));
	set_specular(Color(0.0f));
	set_emission(Color(0.0f));
	set_shininess(50.0f);
	set_reflectivity(0.0f);
}

void BasicMaterial::fill_program_info(string &module_name, map<string, int> &spec_values) const
{
	module_name = "phong.glsl";
	spec_values["use_diffuse_map"] = (diffuse.texture!=0);
	bool use_specular = (specular.texture || specular.value.r || specular.value.g || specular.value.b);
	spec_values["use_specular"] = use_specular;
	spec_values["use_specular_map"] = (specular.texture!=0);
	spec_values["use_shininess_map"] = (use_specular && shininess.texture!=0);
	spec_values["use_normal_map"] = (normal.texture!=0);
	bool use_emission = (emission.texture || emission.value.r || emission.value.g || emission.value.b);
	spec_values["use_emission"] = use_emission;
	spec_values["use_emission_map"] = (emission.texture!=0);
	spec_values["use_reflectivity"] = (reflectivity.value!=0 || reflectivity.texture!=0);
	spec_values["use_reflectivity_map"] = (reflectivity.texture!=0);
	spec_values["use_shadow_map"] = receive_shadows;
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
void BasicMaterial::attach_textures_to(Texturing &texturing, ProgramData &tex_shdata) const
{
	attach_texture_to(diffuse.texture, texturing, tex_shdata, "diffuse_map");
	attach_texture_to(specular.texture, texturing, tex_shdata, "specular_map");
	attach_texture_to(normal.texture, texturing, tex_shdata, "normal_map");
	attach_texture_to(emission.texture, texturing, tex_shdata, "emission_map");
	attach_texture_to(shininess.texture, texturing, tex_shdata, "shininess_map");
	attach_texture_to(reflectivity.texture, texturing, tex_shdata, "reflectivity_map");
}
#pragma GCC diagnostic pop

const Texture *BasicMaterial::get_texture(Tag tag) const
{
	if(tag==texture_tags[0])
		return diffuse.texture;
	else if(tag==texture_tags[1])
		return specular.texture;
	else if(tag==texture_tags[2])
		return shininess.texture;
	else if(tag==texture_tags[3])
		return normal.texture;
	else if(tag==texture_tags[4])
		return emission.texture;
	else if(tag==texture_tags[5])
		return reflectivity.texture;
	else
		return 0;
}

void BasicMaterial::set_diffuse(const Color &color)
{
	diffuse.value = color;
	shdata.uniform("basic_material.diffuse", color);
}

void BasicMaterial::set_diffuse_map(const Texture *tex)
{
	diffuse.texture = tex;
}

void BasicMaterial::set_specular(const Color &color)
{
	specular.value = color;
	shdata.uniform("basic_material.specular", color);
}

void BasicMaterial::set_specular_map(const Texture *tex)
{
	specular.texture = tex;
}

void BasicMaterial::set_normal_map(const Texture *tex)
{
	normal.texture = tex;
}

void BasicMaterial::set_emission(const Color &color)
{
	emission.value = color;
	shdata.uniform("basic_material.emission", color);
}

void BasicMaterial::set_emission_map(const Texture *tex)
{
	emission.texture = tex;
}

void BasicMaterial::set_shininess(float value)
{
	shininess.value = value;
	shdata.uniform("basic_material.shininess", value);
}

void BasicMaterial::set_shininess_map(const Texture *tex)
{
	shininess.texture = tex;
}

void BasicMaterial::set_reflectivity(float value)
{
	reflectivity.value = value;
	shdata.uniform("basic_material.reflectivity", value);
}

void BasicMaterial::set_reflectivity_map(const Texture *tex)
{
	reflectivity.texture = tex;
}

void BasicMaterial::set_receive_shadows(bool s)
{
	receive_shadows = s;
}


DataFile::Loader::ActionMap BasicMaterial::Loader::shared_actions;

BasicMaterial::Loader::Loader(BasicMaterial &m):
	DerivedObjectLoader<BasicMaterial, Material::PropertyLoader<BasicMaterial> >(m)
{
	set_actions(shared_actions);
}

BasicMaterial::Loader::Loader(BasicMaterial &m, Collection &c):
	DerivedObjectLoader<BasicMaterial, Material::PropertyLoader<BasicMaterial> >(m, c)
{
	set_actions(shared_actions);
}

void BasicMaterial::Loader::init_actions()
{
	Material::PropertyLoader<BasicMaterial>::init_actions();
	add_property("diffuse", &BasicMaterial::set_diffuse, &BasicMaterial::set_diffuse_map, true);
	add_property("specular", &BasicMaterial::set_specular, &BasicMaterial::set_specular_map, false);
	add_property("normal", &BasicMaterial::set_normal_map);
	add_property("emission", &BasicMaterial::set_emission, &BasicMaterial::set_emission_map, false);
	add_property("shininess", &BasicMaterial::set_shininess, &BasicMaterial::set_shininess_map);
	add_property("reflectivity", &BasicMaterial::set_reflectivity, &BasicMaterial::set_reflectivity_map);
	add("receive_shadows", &BasicMaterial::receive_shadows);
}

} // namespace GL
} // namespace Msp
