#include "pbrmaterial.h"

using namespace std;

namespace Msp {
namespace GL {

const Tag PbrMaterial::texture_tags[] =
{
	Tag("base_color_map"),
	Tag("normal_map"),
	Tag("metalness_map"),
	Tag("roughness_map"),
	Tag("occlusion_map"),
	Tag("emission_map"),
	Tag()
};

PbrMaterial::PbrMaterial():
	receive_shadows(false)
{
	set_base_color(0.8f);
	set_metalness(0.0f);
	set_roughness(0.5f);
	set_emission(0.0f);
}

void PbrMaterial::fill_program_info(string &module_name, map<string, int> &spec_values) const
{
	module_name = "cooktorrance.glsl";
	spec_values["use_base_color_map"] = (base_color.texture!=0);
	spec_values["use_normal_map"] = (normal.texture!=0);
	spec_values["use_metalness_map"] = (metalness.texture!=0);
	spec_values["use_roughness_map"] = (roughness.texture!=0);
	spec_values["use_occlusion_map"] = (occlusion.texture!=0);
	bool use_emission = (emission.texture || emission.value.r || emission.value.g || emission.value.b);
	spec_values["use_emission"] = use_emission;
	spec_values["use_emission_map"] = (emission.texture!=0);
	spec_values["use_shadow_map"] = receive_shadows;
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
void PbrMaterial::attach_textures_to(Texturing &texturing, ProgramData &tex_shdata) const
{
	attach_texture_to(base_color.texture, texturing, tex_shdata, "base_color_map");
	attach_texture_to(metalness.texture, texturing, tex_shdata, "metalness_map");
	attach_texture_to(roughness.texture, texturing, tex_shdata, "roughness_map");
	attach_texture_to(normal.texture, texturing, tex_shdata, "normal_map");
	attach_texture_to(occlusion.texture, texturing, tex_shdata, "occlusion_map");
	attach_texture_to(emission.texture, texturing, tex_shdata, "emission_map");
}
#pragma GCC diagnostic pop

const Texture *PbrMaterial::get_texture(Tag tag) const
{
	if(tag==texture_tags[0])
		return base_color.texture;
	else if(tag==texture_tags[1])
		return normal.texture;
	else if(tag==texture_tags[2])
		return metalness.texture;
	else if(tag==texture_tags[3])
		return roughness.texture;
	else if(tag==texture_tags[4])
		return occlusion.texture;
	else if(tag==texture_tags[5])
		return emission.texture;
	else
		return 0;
}

void PbrMaterial::set_base_color(const Color &color)
{
	base_color.value = color;
	shdata.uniform("pbr_material.base_color", color);
}

void PbrMaterial::set_base_color_map(const Texture *tex)
{
	base_color.texture = tex;
}

void PbrMaterial::set_normal_map(const Texture *tex)
{
	normal.texture = tex;
}

void PbrMaterial::set_metalness(float value)
{
	metalness.value = value;
	shdata.uniform("pbr_material.metalness", value);
}

void PbrMaterial::set_metalness_map(const Texture *tex)
{
	metalness.texture = tex;
}

void PbrMaterial::set_roughness(float value)
{
	roughness.value = value;
	shdata.uniform("pbr_material.roughness", value);
}

void PbrMaterial::set_roughness_map(const Texture *tex)
{
	roughness.texture = tex;
}

void PbrMaterial::set_occlusion_map(const Texture *tex)
{
	occlusion.texture = tex;
}

void PbrMaterial::set_emission(const Color &color)
{
	emission.value = color;
	shdata.uniform("pbr_material.emission", color);
}

void PbrMaterial::set_emission_map(const Texture *tex)
{
	emission.texture = tex;
}

void PbrMaterial::set_receive_shadows(bool s)
{
	receive_shadows = s;
}


DataFile::Loader::ActionMap PbrMaterial::Loader::shared_actions;

PbrMaterial::Loader::Loader(PbrMaterial &m):
	DerivedObjectLoader<PbrMaterial, Material::PropertyLoader<PbrMaterial> >(m)
{
	set_actions(shared_actions);
}

PbrMaterial::Loader::Loader(PbrMaterial &m, Collection &c):
	DerivedObjectLoader<PbrMaterial, Material::PropertyLoader<PbrMaterial> >(m, c)
{
	set_actions(shared_actions);
}

void PbrMaterial::Loader::init_actions()
{
	Material::PropertyLoader<PbrMaterial>::init_actions();
	add_property("base_color", &PbrMaterial::set_base_color, &PbrMaterial::set_base_color_map, true);
	add_property("normal", &PbrMaterial::set_normal_map);
	add_property("metalness", &PbrMaterial::set_metalness, &PbrMaterial::set_metalness_map);
	add_property("roughness", &PbrMaterial::set_roughness, &PbrMaterial::set_roughness_map);
	add_property("occlusion", &PbrMaterial::set_occlusion_map);
	add_property("emission", &PbrMaterial::set_emission, &PbrMaterial::set_emission_map, false);
	add("receive_shadows", &PbrMaterial::receive_shadows);
}

} // namespace GL
} // namespace Msp
